/*
 *    Copyright (C) 2011 by Sylvain Garsault <sylvain.garsault@gmail.com>    
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Importer.loadQtBinding("qt.core");
Importer.loadQtBinding("qt.gui");
Importer.loadQtBinding("qt.network");
Importer.include("conf.js");
Importer.include("util.js");
Importer.include("amarokContent.js");
Importer.include("amarokControl.js");

/*
 * This class handles an HTTP request, with header and content
 */
function HTTPResponse(code, reason, mimeType)
{
  this.content = new QByteArray();
  this.retCode = code;
  this.reason = reason;
  this.mimeType = mimeType;
  this.header = new QHttpResponseHeader(this.retCode, this.reason, 1, 1);
  this.header.setValue("Content-Type", this.mimeType);
}


HTTPResponse.prototype.append = function(data)
{
  this.content.append(data);
  this.header.setValue("Content-Length", this.content.length());
}


HTTPResponse.prototype.setMimeType = function(type)
{
  this.mimeType = type;
  this.header.setValue("Content-Type", this.mimeType);
}

/*
 * This class creates a TCP socket, able to receive and send data over the network
 */
function TCPSocket(server, socket)
{
  this.server = server;
  this.socket = socket;
  this.id = 0;
  this.socket.readyRead.connect(this, this.onReadyRead);
  this.socket.disconnected.connect(this, this.onDisconnected);
}

TCPSocket.prototype.onReadyRead = function()
{
  d("Ready to read socket " + this.id + " (peer " + this.socket.peerAddress() + ")");
  data = this.socket.readAll().toString();
  d("Request is " + data.length + " bytes long ");
  string = "\r\n\r\n";
  index = data.indexOf(string);
  if (index < 0)
  {
    w("Cannot parse HTTP request from Freebox");
  }
  var header = new QHttpRequestHeader(data);
  this.server.handleRequest(this, header);
  this.socket.close();
}

TCPSocket.prototype.onDisconnected = function()
{
  d("Socket " + this.id + " disconnected");
  this.server.removeSocket(this);
}

TCPSocket.prototype.send = function(response)
{
  data = new QByteArray();
  data.append(response.header.toString());
  // If this is HTML, convert to Freebox charset
  if (response.mimeType=="text/html")
    data.append(u2fbx(response.content));
  else
    data.append(response.content);
  this.socket.write(data);
  d("Socket " + this.id + " sending response, length " + data.length() + " bytes.");
}

TCPSocket.prototype.close = function()
{
  this.socket.close();
}


/*
 * This class creates an HTTP server for the Freebox
 */
function HTTPServer(serverConf)
{
    this.isActive = true;
    // TCP server
    this.TCPServer = new QTcpServer();
    // pool of TCP sockets
    this.sockets = new Array();
    // TCP sessions
    this.sessions = new Array();
    // tag for track changes, in order to refresh Freebox cache
    this.trackTag = (new Date()).getTime();
    // server config
    var serverConf = gConf.server;
    
    //Log some important information for debug
    dlog("=== Starting FreeboKs v" + version(true) +" ===");
    dlog("Using Amarok version " + Amarok.Info.version());
    
    var ret = 0;
    while(!this.TCPServer.isListening())
    {
      // Launch server
      this.TCPServer.listen(new QHostAddress(QHostAddress.Any), serverConf.port);
      switch(ret)
      {
	case 0:
	  ret = 1;
	  break;
	case 2:
	  // end script
	  dlog("=== FreeboKs stopped on user request ===");
	  Amarok.end();
	  break;
	default:
	  e("Unable to connect to port "+ serverConf.port);
	  ret = Amarok.alert("<b>FreeboKs:</b> Unable to open the port "+ serverConf.port +" for the web server.<br />Please check that your firewall is not blocking this port and retry.", "warningContinueCancel"); 
      }
    }
    statusMessage("<b>v" + version() + " succesfully started!</b>  It can be accessed on port "+this.TCPServer.serverPort()+".");
    this.TCPServer.newConnection.connect(this, this.onNewConnection);
    Amarok.Engine.trackChanged.connect(this, this.onTrackChanged);
}

HTTPServer.prototype.onNewConnection = function()
{
  d("===")
  d(this.activeSockets() + " active connection(s)");
  while (this.TCPServer.hasPendingConnections())
  {
    var socket = new TCPSocket(this, this.TCPServer.nextPendingConnection());
    this.addSocket(socket);
    d("Registering new connection");
  }
}

/*
 * Adds a socket to the pool of currently opened sockets
 */
HTTPServer.prototype.addSocket = function(socket)
{
  var i = 0;
  while (i<this.sockets.length && this.sockets[i])
  {
    ++i;
  }
  socket.id = i;
  this.sockets[i]=socket;
  d("Added new socket to the pool["+socket.id+"]");
}

/*
 * Removes a socket from the list of currently opened sockets, removes trailing empty sockets
 */
HTTPServer.prototype.removeSocket = function(socket)
{
  id = socket.id;
  if (this.sockets[id])
  {
    this.sockets[id]=null;
    // remove trailing empty sockets
    var i = this.sockets.length-1;
    var removed = 0;
    while (i>=0 && !this.sockets[i])
    {
      this.sockets.pop();
      --i;
      ++removed;
    }
    d("Deleted socket "+id+" from pool["+this.sockets.length+"]");
    d("Cleaned "+removed+" empty sockets");
  }
  else
  {
    w("Cannot delete socket: not found in pool " + id);
  }
  // delete socket anyway
  delete socket;
}

/*
 * Counts the number of active sockets
 */
HTTPServer.prototype.activeSockets = function()
{
  var numSockets=0;
  for (var i=0; i < this.sockets.length; ++i)
  {
    if (this.sockets[i])
    {
      numSockets++;
    }
  }
  return numSockets;
}

/*
 * Adds a session to the pool of currently opened sessions
 */
HTTPServer.prototype.addSession = function(session)
{
  result=this.sessions.push(session);
  if (result)
  {
    session.id = this.sessions.length;
    d("Added new session to the pool["+session.id+"]");
  }
  else
    w("Could not add session " + session);
  return result;
}

/*
 * Generates a new tag each time a track changes in order to force the Freebox to clear its cache of images
 */
HTTPServer.prototype.onTrackChanged = function()
{
  this.trackTag = (new Date()).getTime();
  d("Track changed, new tag: "+this.trackTag);
}


/*
 * Main function
 */
HTTPServer.prototype.handleRequest = function(socket, requestHeader)
{
  var response = this.parseRequest(socket, requestHeader);
  socket.send(response);
  socket.close();
}

/*
 * Freebox request parser
 */
HTTPServer.prototype.parseRequest = function(socket, requestHeader)
{
  d("Handling request method " + requestHeader.method() + " for URL " +  requestHeader.path());
  var response = new HTTPResponse(200, "OK", "text/html");
  
  if (!(requestHeader.method()=="GET"))
  {
    response.code = 500;
    response.reason = "Expecting a GET request";
    e("Error in request: "+response.reason+", got: "+requestHeader.method());
    return response;
  }
  
  var url = new QUrl(requestHeader.path());
  d("Serving URL "+url);
  path = url.toString().split("/");
  //d("Path["+path.length+"] is "+ path);
  ressource = "";
  root = url.path();
  if (path.length>1)
  {
    resource = path[path.length-1];
  }
  //d("Root is "+root);
  //d("Resource is "+ resource);
  
  // file to server
  file = "";
 // arguments associated to this file
  var arguments = new Array();
  
  if (resource.indexOf("?")<0)
  {
    file = QUrl.fromPercentEncoding(new QByteArray(resource));
  }
  else
  {
    command = resource.split("?");
    //d("Command is "+command);
    
    if (command.length>0)
    file = QUrl.fromPercentEncoding(new QByteArray((command[0])));
    if (command.length > 1)
    {
      d("Parsing an URL containing " + (command.length-1) + " commands.");
      pairs = QUrl.fromPercentEncoding(new QByteArray(command[1])).split(";");
      var i=0;
      for (i=0; i<pairs.length; i++)
      {
	pair = pairs[i].split("=");
	if (pair.length == 2)
	{
	  d("Arg["+i+"]: "+pair);
	  var item = new Object;
	  item.attribute = pair[0];
	  item.value = pair[1];
	  arguments.push(item);
	}
      }
	d("Parsed a command containing " + arguments.length + " attributes.");
     }
   }
  
  realm="";
  if (root.indexOf("/")>=0)
  {
    dirArray = root.split("/");
    if (dirArray.length>2)
      realm = dirArray[1];
    if (dirArray.length>3)
      w("Not expected sub-folders in request " + dirArray);
  }
  else
  {
    realm=root;
  }

  switch(realm)
  {
    case "html":
    case "":
      parser = new HTMLParser(socket, file, arguments);
      response.append(parser.parse());
      break;
    case "img":
      parser = new HTMLParser(socket, file, arguments);
      response.setMimeType("image/png");
      response.append(parser.parseImg(file));
      break;
    default:
      w("Realm error: "+ realm);
  }
  return response;
}




